@@ -4,33 +4,30 @@
 # maps method names to batch functions
 batch_func_map <- list(
   geocodio = batch_geocodio, 
-  census = batch_census
-)
-
-# stores the minimum number of seconds that must elapse for each query
-# based on the usage limit of the service (free tier if there are multiple plans available)
-# Stored value is SECONDS PER QUERY
-usage_limit_map <- list(
-  osm = 1,               # 1 query/second             
-  geocodio = 60/1000,    # 1000 queries per minute (free tier)
-  iq = 1/2,              # 2 queries per second (free tier)
-  google = 1/50          # 50 queries per second
+  census = batch_census,
+  here = batch_here,
+  tomtom = batch_tomtom,
+  mapquest = batch_mapquest,
+  bing = batch_bing
 )
 
 #' Geocode addresses
 #' 
 #' @description
-#' Geocodes addresses given as character values. The \code{\link{geocode}}
+#' Geocodes addresses given as character values. The [geocode]
 #' function utilizes this function on addresses contained in dataframes.
-#' See example usage in \code{vignette("tidygeocoder")} 
+#' See example usage in `vignette("tidygeocoder")`.
 #' 
 #' Note that not all geocoder services support certain address component 
 #' parameters. For example, the Census geocoder only covers the United States 
-#' and does not have a "country" parameter. Refer to \code{\link{api_parameter_reference}} 
-#' for more details on geocoder service parameters and API usage. 
+#' and does not have a "country" parameter. 
 #' 
-#' This function uses the \code{\link{get_api_query}}, \code{\link{query_api}}, and
-#' \code{\link{extract_results}} functions to create, execute, and parse the geocoder
+#' Refer to [api_parameter_reference],
+#' [min_time_reference], and [batch_limit_reference] for more details on 
+#' geocoder service parameters and usage. 
+#' 
+#' This function uses the [get_api_query], [query_api], and
+#' [extract_results] functions to create, execute, and parse geocoder
 #' API queries.
 #' 
 #' @param address single line address (ie. '1600 Pennsylvania Ave NW, Washington, DC').
@@ -43,77 +40,67 @@ usage_limit_map <- list(
 #' @param postalcode postalcode (zip code if in the United States)
 #' @param country country (ie. 'Japan')
 #' 
-#' @param method the geocoder service to be used. Refer to 
-#' \code{\link{api_parameter_reference}} and the API documentation for
-#' each geocoder service for usage details and limitations.
-#' \itemize{
-#'   \item \code{"census"}: US Census Geocoder. US street-level addresses only. 
-#'      Can perform batch geocoding.
-#'   \item \code{"osm"}: Nominatim (OSM). Worldwide coverage.
-#'   \item \code{"geocodio"}: Commercial geocoder. Covers US and Canada and has
-#'      batch geocoding capabilities. Requires an API Key to be stored in
-#'      the "GEOCODIO_API_KEY" environmental variable.
-#'   \item \code{"iq"}: Commercial Nominatim geocoder service. Requires an API Key to
-#'      be stored in the "LOCATIONIQ_API_KEY" environmental variable.
-#'   \item \code{"google"}: Commercial Google geocoder service. Requires an API Key to
-#'      be stored in the "GOOGLEGEOCODE_API_KEY" environmental variable.
-#'   \item \code{"cascade"} : Attempts to use one geocoder service and then uses
+#' @param method `r get_method_documentation(reverse = FALSE)`
+#'  - `"cascade"` : First uses one geocoder service and then uses
 #'     a second geocoder service if the first service didn't return results.
 #'     The services and order is specified by the cascade_order argument. 
-#'     Note that this is not compatible with \code{full_results = TRUE} as geocoder
+#'     Note that this is not compatible with `full_results = TRUE` as geocoder
 #'     services have different columns that they return.
-#' }
+#' 
 #' @param cascade_order a vector with two character values for the method argument 
-#'  in the order in which the geocoder services will be attempted for method = "cascade"
-#'  (ie. \code{c('census', 'geocodio')})
-#' @param lat latitude column name. Can be quoted or unquoted (ie. lat or 'lat').
-#' @param long longitude column name. Can be quoted or unquoted (ie. long or 'long').
-#' @param limit number of results to return per address. Note that not all methods support
-#'  setting limit to a value other than 1. Also limit > 1 is not compatible 
-#'  with batch geocoding if return_addresses = TRUE.
-#' @param min_time minimum amount of time for a query to take (in seconds) if using
-#'  Location IQ or OSM. This parameter is used to abide by API usage limits. You can
-#'  set it to a lower value (ie. 0) if using a local Nominatim server, for instance.
+#'  in the order in which the geocoder services will be attempted for `method = "cascade"`
+#'  (ie. `c("census", "geocodio")`)
+#' @param lat latitude column name. Can be quoted or unquoted (ie. lat or "lat").
+#' @param long longitude column name. Can be quoted or unquoted (ie. long or "long").
+#' @param limit `r get_limit_documentation(reverse = FALSE, df_input = FALSE)`
+#' @param min_time minimum amount of time for a query to take (in seconds). If NULL
+#' then min_time will be set to the default value specified in [min_time_reference].
 #' @param api_url custom API URL. If specified, the default API URL will be overridden.
-#'  This parameter can be used to specify a local Nominatim server.
+#'  This parameter can be used to specify a local Nominatim server, for instance.
 #' @param timeout query timeout (in minutes)
 #' 
-#' @param mode set to 'batch' to force batch geocoding or 'single' to 
-#'  force single address geocoding (one address per query). If not 
-#'  specified then batch geocoding will be used if available
-#'  (given method selected) when multiple addresses are provided, otherwise
-#'  single address geocoding will be used.
+#' @param mode `r get_mode_documentation(reverse = FALSE)`
+
 #' @param full_results returns all data from the geocoder service if TRUE. 
 #' If FALSE then only longitude and latitude are returned from the geocoder service.
-#' @param unique_only only return results for unique addresses if TRUE
+#' @param unique_only only return results for unique inputs if TRUE
 #' @param return_addresses return input addresses with results if TRUE. Note that
-#'    most services return the input addresses with full_results = TRUE and setting
+#'    most services return the input addresses with `full_results = TRUE` and setting
 #'    return_addresses to FALSE does not prevent this.
 #' 
 #' @param flatten if TRUE then any nested dataframes in results are flattened if possible.
-#'    Note that Geocodio batch geocoding results are flattened regardless.
-#' @param batch_limit limit to the number of addresses in a batch geocoding query.
-#'  Both geocodio and census batch geocoders have a 10,000 address limit so this
-#'  is the default.
+#'    Note that in some cases results are flattened regardless such as for
+#'    Geocodio batch geocoding.
+#' @param batch_limit  `r get_batch_limit_documentation(reverse = FALSE)`
+#' @param batch_limit_error `r get_batch_limit_error_documentation(reverse = FALSE)`
 #' @param verbose if TRUE then detailed logs are output to the console
-#' @param no_query if TRUE then no queries are sent to the geocoder and verbose is set to TRUE
+#' @param no_query if TRUE then no queries are sent to the geocoder service and verbose is set to TRUE.
+#'    Used for testing.
 
 #' @param custom_query API-specific parameters to be used, passed as a named list 
-#'  (ie. \code{list(vintage = 'Current_Census2010')}).
-#' @param return_type only used when method = 'census'. Two possible values: 
-#' \itemize{
-#'     \item \code{"locations"} (default)
-#'     \item \code{"geographies"}: returns additional geography columns. 
-#'     See the Census geocoder API documentation for more details.
-#' }
-#' @param iq_region 'us' (default) or 'eu'. Used for establishing API URL for the 'iq' method
-#' @param geocodio_v version of geocodio api. 1.6 is default. Used for establishing API URL
-#'   for the 'geocodio' method.
-#' @param param_error if TRUE then an error will be thrown if certain parameters are invalid for the selected geocoder
-#'   service (method). The parameters checked are limit, address, street, city, county, state, postalcode, and country.
-#'   If method = 'cascade' then no errors will be thrown.
-#' 
-#' @return parsed geocoding results in tibble format
+#'  (ie. `list(extratags = 1)`.
+#' @param return_type only used when `method = "census"`. Two possible values: 
+#'   - `"locations"` (default)
+#'   - `"geographies"`: returns additional geography columns. 
+#'   See the Census geocoder API documentation for more details.
+#' @param iq_region "us" (default) or "eu". Used for establishing the API URL for the "iq" method.
+#' @param geocodio_v version of geocodio API. Used for establishing the API URL
+#'   for the "geocodio" method.
+#' @param param_error if TRUE then an error will be thrown if any address parameters are used that are
+#'   invalid for the selected service (`method`). If `method = "cascade"` then no errors will be thrown.
+#' @param mapbox_permanent if TRUE then the `mapbox.places-permanent`
+#'   endpoint would be used. Note that this option should be used only if you 
+#'   have applied for a permanent account. Unsuccessful requests made by an 
+#'   account that does not have access to the endpoint may be billable.
+#' @param here_request_id This parameter would return a previous HERE batch job,
+#'   identified by its RequestID. The RequestID of a batch job is displayed 
+#'   when `verbose` is TRUE. Note that this option would ignore the 
+#'   current `address` parameter on the request, so `return_addresses`
+#'   needs to be FALSE.
+#' @param mapquest_open if TRUE then MapQuest would use the Open Geocoding 
+#'   endpoint, that relies solely on data contributed to OpenStreetMap.
+#'    
+#' @return tibble (dataframe)
 #' @examples
 #' \donttest{
 #' geo(street = "600 Peachtree Street NE", city = "Atlanta",
@@ -125,16 +112,17 @@ usage_limit_map <- list(
 #' geo(county = 'Jefferson', state = "Kentucky", country = "US",
 #'      method = 'osm')
 #' }
-#' @seealso \code{\link{geocode}} \code{\link{api_parameter_reference}}
+#' @seealso [geocode] [api_parameter_reference] [min_time_reference] [batch_limit_reference]
 #' @export
 geo <- function(address = NULL, 
     street = NULL, city = NULL, county = NULL, state = NULL, postalcode = NULL, country = NULL,
-    method = 'census', cascade_order = c('census', 'osm'), lat = lat, long = long, limit = 1, 
+    method = 'osm', cascade_order = c('census', 'osm'), lat = lat, long = long, limit = 1, 
     min_time = NULL, api_url = NULL, timeout = 20,
     mode = '', full_results = FALSE, unique_only = FALSE, return_addresses = TRUE, 
-    flatten = TRUE, batch_limit = 10000, verbose = FALSE, no_query = FALSE, 
+    flatten = TRUE, batch_limit = NULL, batch_limit_error = TRUE, verbose = FALSE, no_query = FALSE, 
     custom_query = list(), return_type = 'locations', iq_region = 'us', geocodio_v = 1.6, 
-    param_error = TRUE) {
+    param_error = TRUE, mapbox_permanent = FALSE, here_request_id = NULL,
+    mapquest_open = FALSE) {
 
   # NSE - Quote unquoted vars without double quoting quoted vars
   # end result - all of these variables become character values
@@ -145,36 +133,52 @@ geo <- function(address = NULL,
   # IMPORTANT: make sure to put this statement before any other variables are defined in the function
   all_args <- as.list(environment())
   
-  # Reference Variables ------------------------------------------------------------
-  
   # All legal methods (besides 'cascade')
   method_services <- unique(tidygeocoder::api_parameter_reference[['method']])
-  # which methods require an api key
-  methods_requiring_api_key <- unique(tidygeocoder::api_parameter_reference[which(tidygeocoder::api_parameter_reference[['generic_name']] == 'api_key'), ][['method']])
-  
+
   # Check parameter arguments --------------------------------------------------------
-  # all legal method argument excluding 'cascade'
 
   # Check argument inputs
-  stopifnot(is.logical(verbose), is.logical(no_query), is.logical(flatten), is.logical(param_error),
+  check_address_argument_datatype(address, 'address')
+  check_address_argument_datatype(street, 'street')
+  check_address_argument_datatype(city, 'city')
+  check_address_argument_datatype(county, 'county')
+  check_address_argument_datatype(state, 'state')
+  check_address_argument_datatype(postalcode, 'postalcode')
+  check_address_argument_datatype(country, 'country')
+  
+  stopifnot(
+    is.logical(verbose), is.logical(no_query), is.logical(flatten), is.logical(param_error),
             is.logical(full_results), is.logical(unique_only), is.logical(return_addresses),
-            is.numeric(limit), is.numeric(batch_limit), is.numeric(timeout),
-            limit >= 1, batch_limit >= 1, timeout >= 0, is.list(custom_query))
+            is.logical(batch_limit_error), 
+            is.numeric(timeout), timeout >= 0, 
+            is.list(custom_query),
+            is.logical(mapbox_permanent), 
+            is.null(here_request_id) || is.character(here_request_id),
+            is.logical(mapquest_open)
+    )
   
-  if (!(method %in% c('cascade', method_services))) {
-    stop('Invalid method argument. See ?geo')
-  } 
+  check_common_args('geo', mode, limit, batch_limit, min_time)
   
-  if (!(cascade_order[1] %in% method_services) | !(cascade_order[2] %in% method_services) | (length(cascade_order) != 2) |  !(is.character(cascade_order))) {
-    stop('Invalid cascade_order argument. See ?geo')
+  if (mode == 'batch' && (!method %in% names(batch_func_map))) {
+    stop(paste0('The "', method, '" method does not have a batch geocoding function. See ?geo') , call. = FALSE)
   }
   
-  if (!(mode %in% c('', 'single', 'batch'))) {
-    stop('Invalid mode argument. See ?geo')
+  if (!(method %in% c('cascade', method_services))) {
+    stop('Invalid method argument. See ?geo', call. = FALSE)
+  } 
+  
+  if (!(cascade_order[1] %in% method_services) || !(cascade_order[2] %in% method_services) || (length(cascade_order) != 2) || !(is.character(cascade_order))) {
+    stop('Invalid cascade_order argument. See ?geo', call. = FALSE)
+  }
+  
+  if (method == 'cascade' && mode == 'batch' && (length(intersect(cascade_order, names(batch_func_map)) != 2))) {
+    stop("To use method = 'cascade' and mode = 'batch', both methods specified in cascade_order
+          must have batch geocoding capabilities. See ?geo")
   }
   
   if (!(return_type %in% c('geographies', 'locations'))) {
-    stop('Invalid return_type argument. See ?geo')
+    stop('Invalid return_type argument. See ?geo', call. = FALSE)
   }
   
   if (no_query == TRUE) verbose <- TRUE
@@ -186,18 +190,14 @@ geo <- function(address = NULL,
   
   # count number of unique addresses
   num_unique_addresses <- nrow(address_pack$unique) # unique addresses
-  # determine how many rows will be returned (either unique or includes duplicate address rows)
-  # num_rows_to_return <- ifelse(unique_only, num_unique_addresses, nrow(address_pack$crosswalk))
-  #num_rows_to_return <- num_unique_addresses # used for na values
-  
   NA_value <- get_na_value(lat, long, rows = num_unique_addresses) # filler result to return if needed
   
   if (verbose == TRUE) message(paste0('Number of Unique Addresses: ', num_unique_addresses))
   
   # If no valid/non-blank addresses are passed then return NA
-  if (num_unique_addresses == 1 & all(is.na(address_pack$unique))) {
+  if (num_unique_addresses == 1 && all(is.na(address_pack$unique))) {
     if (verbose == TRUE) message(paste0('No non-blank non-NA addreses found. Returning NA results.'))
-    return(unpackage_addresses(address_pack, NA_value, unique_only, return_addresses))
+    return(unpackage_inputs(address_pack, NA_value, unique_only, return_addresses))
   }
   
   ### Parameter Check ------------------------------------------------------
@@ -231,40 +231,37 @@ geo <- function(address = NULL,
                            ' method:\n\n', paste0(illegal_params, collapse = ' '),
                            '\n\nSee ?api_parameter_reference for more details.')
     
-    if (param_error == TRUE) stop(param_message)
+    if (param_error == TRUE) stop(param_message, call. = FALSE)
     else if (verbose == TRUE) message(param_message)
   }
   
   # Cascade method -------------------------------------------------------------
   # If method = 'cascade' is called then pass all function arguments 
-  # except for method and param_error to geo_cascade() and return the results 
+  # except for method, param_error, and batch_limit_error
+  # param_error and batch_limit_error are reverted to FALSE
   if (method == 'cascade') {
-    if (full_results == TRUE) stop("full_results = TRUE cannot be used with the cascade method.")
-    if (limit != 1) stop("limit argument must be 1 (default) to use the cascade method.")
+    if (full_results == TRUE) stop("full_results = TRUE cannot be used with the cascade method.", call. = FALSE)
+    if (is.null(limit) || limit != 1) stop("limit argument must be 1 (default) to use the cascade method.", call. = FALSE)
     
-    return(do.call(geo_cascade, 
-                   c(all_args[!names(all_args) %in% c('method', 'param_error')], list(param_error = FALSE))))
-  }
-  
-  # determine if an illegal limit parameter value was used (ie. if limit !=1
-  # then the method API must have a limit parameter)
-  if ((limit != 1) & (!('limit' %in% legal_parameters))) {
-    illegal_limit_message <- paste0('The limit parameter must be set to 1 (the default) because the "',  method,'" ',
-                                    'method API service does not support a limit argument.\n\n',
-                                    'See ?api_parameter_reference for more details.')
-    
-    if (param_error == TRUE) stop(illegal_limit_message)
-    else if (verbose == TRUE) message(illegal_limit_message)
+    return(do.call(cascade_geocoding, 
+                c(all_args[!names(all_args) %in% c('method', 'param_error', 'batch_limit_error')],
+                list(batch_limit_error = FALSE, param_error = FALSE))))
   }
   
   # Single Address geocoding -------------------------------------------------------------
   # If there are multiple addresses and we are using a method without a batch geocoder function 
   # OR the user has explicitly specified single address geocoding then call the 
   # single address geocoder in a loop (ie. recursively call this function)
-  if ((num_unique_addresses > 1) & ((!(method %in% names(batch_func_map))) | (mode == 'single'))) {
-      if (verbose == TRUE) {
-        message('Executing single address geocoding...\n')
-      }
+  
+  # Exception for geocoder services that should default to single instead of batch
+  if (method %in% pkg.globals$single_first_methods && mode != 'batch' ){
+    mode <- 'single'
+  }
+  
+  # Single address geocoding is used if the method has no batch function or if 
+  # mode = 'single' was specified
+  if ((num_unique_addresses > 1) && ((!(method %in% names(batch_func_map))) || (mode == 'single'))) {
+      if (verbose == TRUE) message('Executing single address geocoding...\n')
       
       # construct arguments for a single address query
       # note that non-address related fields go to the MoreArgs argument of mapply
@@ -281,47 +278,74 @@ geo <- function(address = NULL,
       # rbind the list of tibble dataframes together
       stacked_results <- dplyr::bind_rows(list_coords)
       
-      # note that return_addresses has been set to FALSE here since addresses will already
+      # note that return_inputs has been set to FALSE here since addresses will already
       # be returned in the first geo function call (if asked for)
-      return(unpackage_addresses(address_pack, stacked_results, unique_only, return_addresses = FALSE))
+      return(unpackage_inputs(address_pack, stacked_results, unique_only, FALSE))
       }
       
   # Batch geocoding --------------------------------------------------------------------------
-  if ((num_unique_addresses > 1) | (mode == 'batch')) {
+  if ((num_unique_addresses > 1) || (mode == 'batch')) {
+    if (verbose == TRUE) message('Executing batch geocoding...\n')
     
-    if (limit != 1 & return_addresses == TRUE) {
+    if ((is.null(limit) || limit != 1) && return_addresses == TRUE) {
     stop('For batch geocoding (more than one address per query) the limit argument must 
     be 1 (the default) OR the return_addresses argument must be FALSE. Possible solutions:
     1) Set the mode argument to "single" to force single (not batch) geocoding 
     2) Set limit argument to 1 (ie. 1 result is returned per address)
     3) Set return_addresses to FALSE
-    See the geo() function documentation for details.')
-    }
-  
-    # Enforce batch limit if needed
-    if (num_unique_addresses > batch_limit) {
-      message(paste0('Limiting batch query to ', format(batch_limit, big.mark = ','), ' addresses'))
-      batch_addresses <-  address_pack$unique[1:batch_limit, ]
-    } else {
-      batch_addresses <- address_pack$unique
+    See the geo() function documentation for details.', call. = FALSE)
     }
     
+    # set batch limit to default if not specified
+    if (is.null(batch_limit)) batch_limit <- get_batch_limit(method)
+    if (verbose == TRUE) message(paste0('Batch limit: ', 
+                                        format(batch_limit, big.mark = ',')))
+    
+    # Enforce batch limit if needed
+    if (num_unique_addresses > batch_limit) {
+      batch_limit_exceeded <- TRUE
+      batch_limit_exceeded_message <- paste0(format(num_unique_addresses, big.mark = ','),
+          ' unique addresses found which exceeds the batch limit of ', 
+                                             format(batch_limit, big.mark = ','), '.')
+      
+      if (batch_limit_error == TRUE) stop(batch_limit_exceeded_message, call. = FALSE)
+      else {
+        warning(paste0(batch_limit_exceeded_message, '\n', 
+        'Only geocoding the first ', format(batch_limit, big.mark = ','), ' unique addresses. ',
+                       'All other addresses will have NA results.'))
+        
+        # apply batch limit to query
+        batch_unique_addresses <- address_pack$unique[1 : batch_limit, ]
+      }
+    } else {
+      # if batch limit wasn't exceeded we just will use the unique addresses in address_pack
+      batch_limit_exceeded <- FALSE
+      batch_unique_addresses <- address_pack$unique
+    }
+    
+    # HERE: If a previous job is requested return_addresses should be FALSE
+    # This is because the job won't send the addresses, but would recover the
+    # results of a previous request
+    if (method == 'here' && is.character(here_request_id) && return_addresses == TRUE) {
+      stop('HERE: When requesting a previous job via here_request_id, set return_addresses to FALSE.
+      See the geo() function documentation for details.', call. = FALSE)
+      }
+
     if (verbose == TRUE) message(paste0('Passing ', 
       format(min(batch_limit, num_unique_addresses), big.mark = ','), 
                           ' addresses to the ', method, ' batch geocoder'))
     
-    # Convert our generic query parameters into parameters specific to our API (method)
-    if (no_query == TRUE) return(unpackage_addresses(address_pack, 
-         get_na_value(lat, long, rows = num_unique_addresses), 
-         unique_only, return_addresses))
+    # return NA results if no_query == TRUE
+    if (no_query == TRUE) return(unpackage_inputs(address_pack, NA_value, unique_only, return_addresses))
     
     # call the appropriate function for batch geocoding according the the batch_func_map named list
     # if batch limit was exceeded then apply that limit
-    batch_results <- do.call(batch_func_map[[method]], c(list(batch_addresses),
+    batch_results <- do.call(batch_func_map[[method]], c(list(batch_unique_addresses),
           all_args[!names(all_args) %in% pkg.globals$address_arg_names]))
     
-    # Add NA results if batch limit was reached so rows match up
-    if (num_unique_addresses > batch_limit) {
+    # If batch limit was exceeded we need to add some NA results to make our results
+    # line up with the input addresses
+    if (batch_limit_exceeded == TRUE) {
       batch_filler <- get_na_value(lat, long, rows = num_unique_addresses - batch_limit)
       batch_results <- dplyr::bind_rows(batch_results, batch_filler)
     }
@@ -330,10 +354,11 @@ geo <- function(address = NULL,
     if (verbose == TRUE) {
       batch_time_elapsed <- get_seconds_elapsed(start_time)
       print_time("Query completed in", batch_time_elapsed)
+      message('') # line break
     }
     
     # map the raw results back to the original addresses that were passed if there are duplicates
-    return(unpackage_addresses(address_pack, batch_results, unique_only, return_addresses))
+    return(unpackage_inputs(address_pack, batch_results, unique_only, return_addresses))
   }
 
   #################################################################
@@ -360,59 +385,46 @@ geo <- function(address = NULL,
   
   # Set API URL (if not already set) ---------------------------
   if (is.null(api_url)) {
-    api_url <- switch(method,
-                      "census" = get_census_url(return_type, search),
-                      "osm" = get_osm_url(),
-                      "geocodio" = get_geocodio_url(geocodio_v),
-                      "iq" = get_iq_url(iq_region),
-                      "google" = get_google_url()
-    )
+    api_url <- get_api_url(method, reverse = FALSE, return_type = return_type,
+                search = search, geocodio_v = geocodio_v, iq_region = iq_region, 
+                mapbox_permanent = mapbox_permanent, mapquest_open = mapquest_open)
   }
-  if (length(api_url) == 0) stop('API URL not found')
   
-  # Set min_time if not set based on usage limit of service
-  if (is.null(min_time)) {
-    if (method %in% names(usage_limit_map)) min_time <- usage_limit_map[[method]]
-    else min_time <- 0 # default to 0
+  # Workaround for Mapbox/TomTom - The search_text should be in the API URL
+  if (method %in% c('mapbox', 'tomtom')) {
+    api_url <- gsub(" ", "%20", paste0(api_url, generic_query[['address']], ".json"))
+    # Remove semicolons (Reserved for batch)
+    api_url <- gsub(";", ",", api_url)
   }
 
-  # If API key is required then use the get_key() function to retrieve it
-  if (method %in% methods_requiring_api_key) {
-    generic_query[['api_key']] <- get_key(method)
-  }
-  
-  if (!is.null(limit)) generic_query[['limit']] <- limit
+  # Set min_time if not set based on usage limit of service
+  if (is.null(min_time)) min_time <- get_min_query_time(method)
+
+  # add limit and api_key to generic query
+  generic_query <- add_common_generic_parameters(generic_query, method, no_query, limit)
   
   # Convert our generic query parameters into parameters specific to our API (method)
   api_query_parameters <- get_api_query(method, generic_query, custom_query)
   
   # Execute Single Address Query -----------------------------------------
   if (verbose == TRUE) display_query(api_url, api_query_parameters)
-  if (no_query == TRUE) return(unpackage_addresses(address_pack, NA_value, unique_only, return_addresses))
-  raw_results <- jsonlite::fromJSON(query_api(api_url, api_query_parameters))
   
+  # return NA results if no_query = TRUE
+  if (no_query == TRUE) return(unpackage_inputs(address_pack, NA_value, unique_only, return_addresses))
+  query_results <- query_api(api_url, api_query_parameters, method = method)
+  
+  if (verbose == TRUE) message(paste0('HTTP Status Code: ', as.character(query_results$status)))
   
   ## Extract results -----------------------------------------------------------------------------------
-  # output error message for geocodio if present
-  if ((method == 'geocodio') & (!is.data.frame(raw_results)) & ("error" %in% names(raw_results))) {
-    message(paste0('Error: ', raw_results$error))
+  # if there were problems with the results then return NA
+  if (query_results$status != 200) {
+    extract_errors_from_results(method, query_results$content, verbose)
     results <- NA_value
-  } 
-  # output error message for google if present
-  else if ((method == 'google') & (!is.data.frame(raw_results)) & ("error_message" %in% names(raw_results))) {
-    message(paste0('Error: ', raw_results$error_message))
-    results <- NA_value
-  } 
-  else if (length(raw_results) == 0) {
-    # If no results found, return NA
-    # otherwise extract results
-    results <- NA_value
-    if (verbose == TRUE) message("No results found")
-  } 
+  }
   else {
     # Extract results. Use the full_results and flatten parameters
     # to control the output
-    results <- extract_results(method, raw_results, full_results, flatten)
+    results <- extract_results(method, jsonlite::fromJSON(query_results$content), full_results, flatten, limit)
     
     # Name the latitude and longitude columns in accordance with lat/long arguments
     names(results)[1] <- lat
@@ -421,7 +433,7 @@ geo <- function(address = NULL,
   
   # Make sure the proper amount of time has elapsed for the query per min_time
   pause_until(start_time, min_time, debug = verbose) 
-  if (verbose == TRUE) message() # insert ending line break if verbose
+  if (verbose == TRUE) message('') # insert ending line break if verbose
   
-  return(unpackage_addresses(address_pack, results, unique_only, return_addresses))
+  return(unpackage_inputs(address_pack, results, unique_only, return_addresses))
 }
